/*
 * Copyright (C) 1997-2003 LAAS/CNRS 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */
/*
 *
 * Maher Khatib - Thu Jun 19 1997
 *
 */
#include <h2mathLib.h>
#include <stdio.h>
#include <stdlib.h>
#include <h2errorLib.h>
#include <string.h>
#include <stdarg.h>
#include <posterLib.h>
#include <errnoLib.h>
#include <taskLib.h>

#include <sys/time.h>
#include <portLib.h>

#ifdef __sun
#define HAS_POSIX_CLOCK
#endif
#ifdef __linux
#undef HAS_POSIX_CLOCK
#endif

/*
 * Sick 
 */
#include "sick_lib_local_str.h"
#include "sick_lib.h"

static unsigned short build_crc16 ( char *buf, unsigned short len );
static STATUS send_control_command ( SICK_LIB_STR *sick, char cmd );
static STATUS send_general_command ( SICK_LIB_STR *sick, char cmd, char *data, unsigned short n );
static STATUS send_mode_command ( SICK_LIB_STR *sick, char mode );
static STATUS sick_lib_check_command ( char cmd );
static char * sick_lib_find_tram_header ( char *buf, int size );
static char * sick_lib_find_tram_header_3 ( char *buf, int size );
static long sick_lib_parse_tram ( SICK_LIB_DATA_TRAM_STR *tram, int n, int *state );
static void sick_lib_update_measure ( SICK_LIB_STR *sick, double *d, double *theta );
static STATUS sick_lib_write_poster ( SICK_LIB_DATA_TRAM_STR *tram );


SERIAL_BUFFER_STR *serial = NULL;

static void
show_cmd(char *who, unsigned char cmd, char *data, int n);
static int
decode_status(unsigned char status, 
	      int *implausible_measure, int *pollution, char *statusString);

/*
 * Initialize sick serial communications and sensor
 */
SICK_LIB_STR *
sick_lib_init (char *device, int rate)
{
  SICK_LIB_STR *sick;

  /* allocation */
  if (!(sick = (SICK_LIB_STR *) malloc (sizeof(SICK_LIB_STR))))
    return NULL;

  /* Init serial port */
  sick->serial = serial_init (device, rate, 
			      SICK_LIB_SERIAL_IN_BUFFER_LENGTH, 
			      SICK_LIB_SERIAL_OUT_BUFFER_LENGTH, 
			      SICK_LIB_SERIAL_READ_TASK_PRIORITY,
			      SICK_LIB_SERIAL_WRITE_TASK_PRIORITY);

  /* check it */
  if (!sick->serial) {
    fprintf (stderr, "sick_lib_init: Cannot initialize serial port!\n");
    h2printErrno (errnoGet());
    return NULL;
  }
  
  /* init state */
  memset (&sick->state, 0, sizeof (SICK_LIB_STATE_STR));
  sick->state.run = SICK_LIB_OK;
  sick->state.start = sick_lib_mtime ();

  /* received fields */
  sick->rcv.mode = SICK_LIB_ONE_CUT_MODE;
  sick->rcv.cmd = SICK_LIB_NO_CMD;
  sick->rcv.range = SICK_LIB_SCAN_RANGE_180;
  sick->rcv.resolution = SICK_LIB_SCAN_RESOLUTION_50;
  sick->rcv.baud = SICK_LIB_DEFAULT_TR_RATE;
  sick->rcv.measure_unit = SICK_LIB_MEASURE_UNIT_MM;

  /* XXX default values compatible for SICK2 
     (because this sick is not controlable) */
  sick->rcv.measure_unit = SICK_LIB_MEASURE_UNIT_CM;
  sick->config.measure_flag = SICK_LIB_MEASURE_FLAGS_A;
  sick->rcv.resolution = SICK_LIB_SCAN_RESOLUTION_100;

  /* same as send */
  sick->send = sick->rcv;

  /* Init of points */
  sick->measures.n = 0;

  /* OK */
  return sick;
}

/*
 * Free sick serial communications objects
 */

void
sick_lib_end (SICK_LIB_STR *sick)
{
  /* check */
  if (!sick) 
    return;

  /* free serial control */
  if (sick->serial)
    serial_free (sick->serial);
  
  /* free it */
  free (sick);
}

/* 
 * Send reset command 
 *
 * return OK / ERROR.
 */

STATUS 
sick_lib_reset_cmd (SICK_LIB_STR *sick)
{
  /* init reset */
  sick->state.reset = FALSE;
  sick->send.cmd = INIT_TGM;
  sick->rcv.cmd = SICK_LIB_NO_CMD;

  /* send command */
  return send_control_command (sick, INIT_TGM);
}


/*
 * Get sick type
 */
STATUS
sick_lib_get_type_cmd (SICK_LIB_STR *sick)
{
  sick->send.cmd = TYPANF_TGM;
  sick->rcv.cmd = SICK_LIB_NO_CMD;
  
  /* send it */
  return send_control_command (sick, TYPANF_TGM);
}

/* 
 * Send read config
 *
 * return OK / ERROR.
 */

STATUS 
sick_lib_read_config_cmd (SICK_LIB_STR *sick)
{
  /* init config flag */
  sick->state.config = FALSE;
  sick->send.cmd = LMSKFGANF_TGM;
  sick->rcv.cmd = SICK_LIB_NO_CMD;
  
  /* send command */
  return send_control_command (sick, LMSKFGANF_TGM);
}

/* 
 * Send set config
 *
 * return OK / ERROR.
 */

STATUS 
sick_lib_set_config_cmd (SICK_LIB_STR *sick)
{
  unsigned char mode = sick->rcv.mode;

  /* check sick state */
  if (mode != SICK_LIB_CONFIG_MODE || !sick->state.config) {
    errnoSet (S_sickLib_bad_mode);
    return ERROR;
  }

  sick->send.cmd = LMSKFGDEF_TGM;
  sick->rcv.cmd = SICK_LIB_NO_CMD;

  /* send changed config */
  if (send_general_command (sick, LMSKFGDEF_TGM, (char *)&sick->config,
			    sizeof (SICK_LIB_INTERNAL_CONFIG_STR)) != OK) {
    return ERROR;
  }

  return OK;
}

/* 
 * Send Get Status command 
 *
 * return OK / ERROR.
 */

STATUS 
sick_lib_read_state_cmd (SICK_LIB_STR *sick)
{
  /* init of state flag */
  sick->state.state = FALSE;

  sick->send.cmd = SSANF_TGM;
  sick->rcv.cmd = SICK_LIB_NO_CMD;

  /* send command */
  return send_control_command (sick, SSANF_TGM);
}

/* 
 * Send Get Status command 
 *
 * return OK / ERROR.
 */

STATUS 
sick_lib_get_measures_cmd (SICK_LIB_STR *sick)
{
  char measure_type = 1;

  sick->send.cmd = MWANF_TGM;
  sick->rcv.cmd = SICK_LIB_NO_CMD;

  if (send_general_command (sick, MWANF_TGM, &measure_type, 1) != OK)
    return ERROR;

  /* time */
  sick->state.start = sick_lib_mtime ();

  return OK;
}

/* 
 * Send One Cut mode
 *
 * return OK / ERROR.
 */

STATUS 
sick_lib_one_cut_mode (SICK_LIB_STR *sick)
{
  sick->send.cmd = BM_TGM;
  sick->rcv.cmd = SICK_LIB_NO_CMD;

  if (send_mode_command (sick, SICK_LIB_ONE_CUT_MODE) != OK)
    return ERROR;

  /* time */
  sick->state.start = sick_lib_mtime ();
  return OK;
}

/* 
 * Send measures continuously mode
 *
 * return OK / ERROR.
 */

STATUS 
sick_lib_cont_cut_mode (SICK_LIB_STR *sick)
{
  sick->send.cmd = BM_TGM;
  sick->rcv.cmd = SICK_LIB_NO_CMD;

  if (send_mode_command (sick, SICK_LIB_CONT_CUT_MODE) != OK)
    return ERROR;

  /* get time */
  sick->state.start = sick_lib_mtime ();

  return OK;
}

/* 
 * Send config mode
 *
 * return OK / ERROR.
 */

STATUS 
sick_lib_config_mode (SICK_LIB_STR *sick)
{
  char buf[10];

  /* bulid buffer */
  buf[0] = SICK_LIB_CONFIG_MODE;
  strcpy (&buf[1], SICK_LIB_PASSWD);

  sick->send.cmd = BM_TGM;
  sick->rcv.cmd = SICK_LIB_NO_CMD;
  sick->send.mode = (int) SICK_LIB_CONFIG_MODE;

  /* send it */
  return send_general_command (sick, BM_TGM, buf, strlen (SICK_LIB_PASSWD) + 1);
}


/*
 * Sick set scan params.
 */
STATUS 
sick_lib_set_scan_params_cmd (SICK_LIB_STR *sick)
{
  char buf[4];
  unsigned short *p;

  /* build buffer */
  p = (unsigned short *) buf;

  /* set range */
  *p = SICK_LIB_SHORT_FORMAT ((unsigned short)sick->send.range);

  p++;

  /* set resolution */
  *p = SICK_LIB_SHORT_FORMAT((unsigned short) sick->send.resolution);

  sick->send.cmd = VARDEF_TGM;
  sick->rcv.cmd = SICK_LIB_NO_CMD;

  /* send it */
  return send_general_command (sick, VARDEF_TGM, buf, 4);
}


/*
 * Sick change baud rate
 */
STATUS
sick_lib_set_baud_rate_cmd (SICK_LIB_STR *sick)
{
  char mode;

  /* check current sick mode */
  if (!sick->state.config || sick->rcv.mode != SICK_LIB_CONFIG_MODE) {
    errnoSet (S_sickLib_bad_mode);
    return ERROR;
  }

  switch (sick->send.baud) {
  case 9600:
    mode = SICK_LIB_9600_RATE_MODE;
    break;
  case 19200:
    mode = SICK_LIB_19200_RATE_MODE;
    break;
  case 38400:
    mode = SICK_LIB_38400_RATE_MODE;
    break;
  default:
    return ERROR;
  } /* switch */

  sick->send.cmd = mode;
  sick->rcv.cmd = SICK_LIB_NO_CMD;

  /* send it */
  if (send_mode_command (sick, mode) != OK)
    return ERROR;

  return OK;
}


/*
 * Sick check if the received cmd correspond to the sent one 
 */
BOOL
sick_lib_is_good_ack_cmd (SICK_LIB_STR *sick)
{
  char s = sick->send.cmd;
  char r = sick->rcv.cmd;

  /* check it */
  if ((s == (r & 0x7f)) || 
      (s == SICK_LIB_9600_RATE_MODE && r == BMACK_TGM) ||
      (s == SICK_LIB_19200_RATE_MODE && r == BMACK_TGM) ||
      (s == SICK_LIB_38400_RATE_MODE && r == BMACK_TGM)
      )
    return TRUE;

  return FALSE;
}

/*
 * Set measure unit
 */
STATUS 
sick_lib_set_measure_unit (SICK_LIB_STR *sick)
{
  /* set it in config str */
  sick->config.measure_unit = sick->send.measure_unit;
  /* set the desired measure flags XXX only possible with 1.07 firmware */
  sick->config.measure_flag = SICK_LIB_MEASURE_FLAGS_A;

  /* call config set function */
  return sick_lib_set_config_cmd (sick);
}

/*
 * Get errors
 */
STATUS 
sick_lib_get_error_cmd (SICK_LIB_STR *sick)
{
  sick->send.cmd = ERRANF_TGM;
  sick->rcv.cmd = SICK_LIB_NO_CMD;
  /* send it */
  return send_control_command (sick, ERRANF_TGM);
}


/*
 * check measure tram
 */
BOOL
sick_lib_is_measure_tram (SICK_LIB_STR *sick)
{
  return sick->tram.cmd == MW_TGM;
}


/*
 * Read last complet tram data. All previous one are removed.
 */

int
sick_lib_read_last_entry_tram (SICK_LIB_STR *sick)
{
  static SICK_LIB_DATA_TRAM_STR tramTmp;
  int size, sizeTmp=0;
  int removed=0;
  int pollution, implausible_measure, status;
  char statusStr[16];

  /* Read all trams until buffer empty */
  while((size=sick_lib_read_entry_tram(sick)) != 0) {
    memcpy(&tramTmp, &sick->tram, sizeof(SICK_LIB_DATA_TRAM_STR));
    sizeTmp = size;
    removed++;
  }
  
  /* Copy last good tram if any */
  if (sizeTmp)
    memcpy(&sick->tram, &tramTmp, sizeof(SICK_LIB_DATA_TRAM_STR));

  if (removed>1)
    logMsg ("Sick lost %d tram\n", removed-1);

  status = decode_status(sick->tram.status, &implausible_measure, 
			 &pollution, statusStr);

  return sizeTmp;
}

/*
 * Read tram data
 */

int
sick_lib_read_entry_tram (SICK_LIB_STR *sick)
{
  int n, size;
  STATUS state;

  /* get serial in buffer length */
  n = serial_in_get_n_chars(sick->serial);

  /* Impose minimum tram length */
  if (n <= SICK_LIB_CONTROL_TRAM_LENGTH) 
    return 0;
  
  /* read buffer it without scrolling */
  n = serial_read_buffer_without_scroll (sick->serial, 
					 sick->tram.data, 
					 SICK_LIB_DATA_LENGTH_MAX);

  /* check it */
  if (n < 1) 
    return ERROR;

  /* parse tram -- compute len, command and data */
  size = sick_lib_parse_tram (&sick->tram, n, &state);

  /* check size <-> n */
  if (size > n) {
    fprintf (stderr, "\t ->Error in Parser size = %d -- n = %d\n", size, n);
    size = n;
  }

  /* check size */
  if (size == 0)
    return 0;

  /* Read input buffer */
  if (serial_scroll_in_buffer (sick->serial, size) != size) 
    return ERROR;
  
  return (state == OK) ? size : 0;
}

/*
 * update sick str
 */
STATUS
sick_lib_update_sick_str (SICK_LIB_STR *sick, double *d, double *theta)
{
  static SICK_LIB_INTERNAL_STATUS_STR sick_lib_status;

  /* update command */
  sick->rcv.cmd = (int)sick->tram.cmd;
  /* suppose it good */
  sick->rcv.result = TRUE;
  /* update sick running status */
  sick->state.run = (SICK_LIB_RUNNING_STATE) sick->tram.status;

  switch (sick->tram.cmd) {
  case MW_TGM:			/* Measurements */
    /* Timing */
    sick->state.end = sick_lib_mtime ();
    /* update measures */
    sick_lib_update_measure (sick, d, theta);
    break;
  case PON_TGM:			/* Power On tram */
    sick->state.reset = TRUE;
    break;
  case INIT_ACK_TGM:		/* Init response tram */
    break;
  case SS_TGM:			/* read state response tram */
    /* copy data in local str */
    memcpy ((void *) &sick_lib_status, sick->tram.data, sick->tram.len);
    /* signal this reading */
    sick->state.state = TRUE;
    break;
  case BMACK_TGM:		/* change operating mode tram */
    /* check it */
    if (*sick->tram.data)
      sick->rcv.result = FALSE;
    /* update mode */
    sick->rcv.mode = sick->send.mode; 
    sick->rcv.baud = sick->send.baud;
    break;
  case NACK_TGM:		/* Error tram */
    /* set error */
    sick->rcv.result = FALSE;
    break;
  case LMSKFG_TGM:		/* read config tram */
    sick->state.config = TRUE;
    memcpy ((void *) &sick->config, sick->tram.data, sick->tram.len);
    break;
  case LMSKFGACK_TGM:		/* read config response tram */
    /* check it */
    if ((*sick->tram.data)) {
      sick->rcv.result = TRUE;
      /* set measures unit */
      sick->rcv.measure_unit = sick->send.measure_unit;
    } else
      /* error */
      sick->rcv.result = FALSE;
    break;
  case VARDEFACK_TGM:		/* change scan value tram */
    if (! (*sick->tram.data)) 
      sick->rcv.result = FALSE;
    else {
      sick->rcv.range = sick->send.range;
      sick->rcv.resolution = sick->send.resolution;
    }
    break;
  case ERR_TGM:			/* Get Errors */
    /* sick->state.run = (SICK_LIB_RUNNING_STATE) (*sick->tram.data & 0x7f); */
    break;
  case TYPE_TGM:
    sick->tram.data[sick->tram.len] = 0;
    strncpy (sick->state.type, sick->tram.data, SICK_LIB_TYPE_LENGTH_MAX - 1);
    sick->state.type[SICK_LIB_TYPE_LENGTH_MAX - 1] = 0;
    break;
  default:
    fprintf(stderr, "unimplemented command: 0x%2x\n", (int)sick->tram.cmd);
  }

  /* return result */
  return (sick->rcv.result) ? OK : ERROR;
}


int
sick_lib_get_length_range(SICK_LIB_STR *sick)
{
  switch(sick->config.measure_unit) {

  case SICK_LIB_MEASURE_UNIT_MM:

    switch(sick->config.measure_flag) {
    case SICK_LIB_MEASURE_FLAGS_A:
    case SICK_LIB_MEASURE_FLAGS_REF2:
      return SICK_LIB_LENGTH_RANGE_32_M;
    case SICK_LIB_MEASURE_FLAGS_AB:
    case SICK_LIB_MEASURE_FLAGS_REF4:
      return SICK_LIB_LENGTH_RANGE_16_M;
    default:
      return SICK_LIB_LENGTH_RANGE_8_M;
    }

  case SICK_LIB_MEASURE_UNIT_CM:

    switch(sick->config.measure_flag) {
    case SICK_LIB_MEASURE_FLAGS_A:
    case SICK_LIB_MEASURE_FLAGS_REF2:
      return SICK_LIB_LENGTH_RANGE_320_M;
    case SICK_LIB_MEASURE_FLAGS_AB:
    case SICK_LIB_MEASURE_FLAGS_REF4:
      return SICK_LIB_LENGTH_RANGE_160_M;
    default:
      return SICK_LIB_LENGTH_RANGE_80_M;
    }

  case SICK_LIB_MEASURE_UNIT_10_CM:

    return SICK_LIB_LENGTH_RANGE_150_M;

  }

  return -1;
}


/********************************************************************************/

/*
 * Static functions
 */

/*
 * Send a control request 
 */

static STATUS
send_control_command (SICK_LIB_STR *sick, char cmd)
{
  SICK_LIB_CONTROL_TRAM_STR control_tram = SICK_LIB_CONTROL_TRAM_DEFAULT;

  show_cmd("CNTL", (unsigned char)cmd, NULL, 0);

  /* build tram */
  control_tram.len = SICK_LIB_SHORT_FORMAT(1);
  control_tram.cmd = cmd;
  /* compute crc */
  control_tram.crc = build_crc16 ((char *)&control_tram, 
				  SICK_LIB_CONTROL_TRAM_LENGTH - 
				  sizeof(unsigned short));
  /* send it */
  if (serial_write_to_buffer (sick->serial, (char *)&control_tram, 
			      SICK_LIB_CONTROL_TRAM_LENGTH) !=
      SICK_LIB_CONTROL_TRAM_LENGTH) {
    fprintf(stderr, "send_control_command: Cannot send tram!\n");
    return ERROR;
  }
  return OK;
}

/*
 * Send a mode request 
 */

static STATUS
send_mode_command (SICK_LIB_STR *sick, char mode)
{
  SICK_LIB_MODE_TRAM_STR mode_tram = SICK_LIB_MODE_TRAM_DEFAULT;

  show_cmd("MODE", (unsigned char)mode, NULL, 0);

  /* build tram */
  mode_tram.len = SICK_LIB_SHORT_FORMAT(2);
  mode_tram.cmd = BM_TGM;
  /* set mode */
  mode_tram.mode = mode;
  sick->send.mode = (int) mode;

  /* compute crc */
  mode_tram.crc = build_crc16 ((char *)&mode_tram, 
			       SICK_LIB_MODE_TRAM_LENGTH - sizeof(unsigned short));

  /* send it */
  if (serial_write_to_buffer (sick->serial, (char *)&mode_tram, 
			      SICK_LIB_MODE_TRAM_LENGTH) != 
      SICK_LIB_MODE_TRAM_LENGTH) {
    fprintf(stderr, "send_mode_command: Cannot send tram!\n");
    return ERROR;
  }

  return OK;
}

/* ----------------------------------------------------------------- */
static void
show_cmd(char *who, unsigned char cmd, char *data, int n)
{
  int i;
  char cmdname[64]= "\0";

  switch(cmd) {
  case INIT_TGM: break;
  case BM_TGM: strcpy(cmdname, "BM_TGM"); break;
  case MWANF_TGM: strcpy(cmdname, "MWANF_TGM"); break;
  case SSANF_TGM: strcpy(cmdname, "SSANF_TGM"); break;
  case ERRANF_TGM: strcpy(cmdname, "ERRANF_TGM"); break;
  case TSTANF_TGM: strcpy(cmdname, "TSTANF_TGM"); break;
  case MEMANF_TGM: strcpy(cmdname, "MEMANF_TGM"); break;
  case MMWANF_TGM: strcpy(cmdname, "MMWANF_TGM"); break;
  case MWPANF_TGM: strcpy(cmdname, "MWPANF_TGM"); break;
  case BSFKFG_TGM: strcpy(cmdname, "BSFKFG_TGM"); break;
  case BSFCONF_TGM: strcpy(cmdname, "BSFCONF_TGM"); break;
  case PWCHG_TGM: strcpy(cmdname, "PWCHG_TGM"); break;
  case KSFKFG_TGM: strcpy(cmdname, "KSFKFG_TGM"); break;
  case WANL_TGM: strcpy(cmdname, "WANL_TGM"); break;
  case SFANF_TGM: strcpy(cmdname, "SFANF_TGM"); break;
  case LMOD_TGM: strcpy(cmdname, "LMOD_TGM"); break;
  case WANLCFRM_TGM: strcpy(cmdname, "WANLCFRM_TGM"); break;
  case AGL_TGM: strcpy(cmdname, "AGL_TGM"); break;
  case ANLDEF_TGM: strcpy(cmdname, "ANLDEF_TGM"); break;
  case ANLCFRM_TGM: strcpy(cmdname, "ANLCFRM_TGM"); break;
  case OUTPUTDEF_TGM: strcpy(cmdname, "OUTPUTDEF_TGM"); break;
  case RSPTIMEDEF_TGM: strcpy(cmdname, "RSPTIMEDEF_TGM"); break;
  case RSPTIMECFRM_TGM: strcpy(cmdname, "RSPTIMECFRM_TGM"); break;
  case WSFTYPDEF_TGM: strcpy(cmdname, "WSFTYPDEF_TGM"); break;
  case BRPERMDEF_TGM: strcpy(cmdname, "BRPERMDEF_TGM"); break;
  case ADRDEF_TGM: strcpy(cmdname, "ADRDEF_TGM"); break;
  case LASER_TGM: strcpy(cmdname, "LASER_TGM"); break;
  case LMSKFGANF_TGM: strcpy(cmdname, "LMSKFGANF_TGM"); break;
  case LMSKFGDEF_TGM: strcpy(cmdname, "LMSKFGDEF_TGM"); break;
  case TYPANF_TGM: strcpy(cmdname, "TYPANF_TGM"); break;
  case VARDEF_TGM: strcpy(cmdname, "VARDEF_TGM"); break;
  default: strcpy(cmdname, "UNKNOWN CMD"); break;
  }

  printf ("sick: send %s cmd 0x%x %s(", who, cmd, cmdname);
  for(i=0;i<n;i++) {
    printf ("%x ", (unsigned char)data[i]);
  }
  printf (")\n");
}

/*
 * Send general command 
 */

static STATUS
send_general_command (SICK_LIB_STR *sick, char cmd, char *data, unsigned short n)
{
  static char buf[SICK_LIB_GEN_CMD_LENGTH_MAX];
  char *p = buf;
  
  show_cmd("GENERAL", (unsigned char)cmd, data, n);

  /* reset */
  if (!data) 
    n = 0;

  /* header */
  *p = SICK_LIB_STX; p++;
  *p = SICK_LIB_ADR; p++;

  /* len */
  *((unsigned short *) p) = SICK_LIB_SHORT_FORMAT(n + 1); 
  p += sizeof(unsigned short);

  /* command */
  *p = cmd; p++;

  /* check data */
  if (data) {
    /* copy data */
    memcpy (p, data, n); p += n;
  }

  /* compute CRC */
  *((unsigned short *) p) = build_crc16 (buf, n + 5);

  /* send it */
  if (serial_write_to_buffer (sick->serial, buf, n + 7) != n + 7) {
    fprintf(stderr, "send_general_command: Cannot send tram!\n");
    return ERROR;
  }

  /* ok */
  return OK;
}

/*
 * Compute CRC16 test code
 */
#ifndef __i386
static unsigned short
build_crc16 (char *buf, unsigned short len)
{
  unsigned short crc16 = 0;
  char data[2];
  
  data [1] = 0;

  while (len--) {
    data[0] = data[1];
    data[1] = *buf++;
    if (crc16 & 0x8000) {
      crc16 <<= 1;
      crc16 ^=  CRC16_GEN_POLY;
    } else
      crc16 <<= 1;

    crc16 ^=  *(unsigned short *)data;
  } /* while */

  return SICK_LIB_SHORT_FORMAT(crc16);
}
#else
static unsigned short
build_crc16 (char *buf, unsigned short len)
{
  unsigned short uChecksum =0;
  char abData[2];
	
  abData[0] =0;
  while(len--)
    {
      abData[1] =abData[0];
      abData[0] =*buf++;
      if (uChecksum & 0x8000)
	{
	  uChecksum <<=1;
	  uChecksum ^=CRC16_GEN_POLY;
	}
      else uChecksum <<=1;
      
      /* ##### This operation probably isn't byte-order safe */
      uChecksum ^=*(unsigned int *)&abData;
    }
  return(uChecksum);
}
#endif

/*
 * Check command
 */
static STATUS
sick_lib_check_command (char cmd)
{
  static char cmds[] = SICK_LIB_TO_STATION_COMMAND_LIST;
  int i;

  /* for all commands */
  for (i = 0; i < SICK_LIB_TO_STATION_COMMAND_LIST_NUMBER; i++) {
    if (cmd == cmds[i]) return OK;
  }
  logMsg ("Sick : unimplemented command: 0x%x\n", (int)cmd);
  return ERROR;
}


/*
 * Get time in milliseconds
 */
unsigned int
sick_lib_mtime (void)
{
  unsigned int m;
#ifdef HAS_POSIX_CLOCK  
  struct timespec tp;
  if (clock_gettime(CLOCK_REALTIME, &tp) != 0)
    return 0;
  m = (tp.tv_nsec / 1000000) + (tp.tv_sec * 1000);
#else  
  struct timeval tv;
  if (gettimeofday(&tv, NULL) != 0) {
      return 0;
  }
  m = (tv.tv_usec/1000) + (tv.tv_sec * 1000);
#endif
  return m;
}

/*
 * decode_status 
 *
 * return 0 if fatal error
 */
static int
decode_status(unsigned char status, 
	      int *implausible_measure, int *pollution, char *statusString) 
{
  unsigned char s012 = status & 0x7;
  unsigned char s6 = status & 0x40;
  unsigned char s7 = status & 0x80;
  
  *implausible_measure = s6;
  if (s6) 
    printf ("sick : IMPLAUSIBLE_MEASURE\n");

  *pollution = s7;
  if (s7)
    printf ("sick : POLLUTION\n");

  if (s012==0) {
    strcpy(statusString, "OK");
    return 1;
  }
  if (s012==1) {
    strcpy(statusString, "info");
    return 1;
  }
  if (s012==2) {
    strcpy(statusString, "warning");
    return 1;
  }
  if (s012==3) {
    strcpy(statusString, "error");
    return 1;
  }
  if (s012==4) {
    strcpy(statusString, "fatal");
    printf ("sick : FATAL ERROR\n");
    return 0;
  }
  return 0;
}

/*
 * Show command str
 */
void
sick_lib_tram_show (SICK_LIB_DATA_TRAM_STR *tram)
{
  unsigned short np;
  char *p = tram->data;
  int pollution, implausible_measure, status;
  char statusStr[16];


  status = decode_status(tram->status, &implausible_measure, &pollution, statusStr);

  printf ("sick: rcv Len %d -- status %s (%d) %s%s -- ", 
	  tram->len, statusStr, tram->status, 
	  implausible_measure?" IMPLAUSIBLE ":"", 
	  pollution?" POLLUTION ":"");

  switch (tram->cmd) {
  case MW_TGM:			/* Measurements */
    /* Get number */
    np = SICK_LIB_SHORT_FORMAT(*((unsigned short *) p)) & 0x0fff;
    printf("MW_TGM Read measures command - %d Points\n", np);
    break;
  case PON_TGM:			/* Power On tram */
    printf("PON_TGM Read Power On TGM. -- ");
    tram->data[tram->len] = 0;
    puts (tram->data);
    break;
  case INIT_ACK_TGM:		/* Init response tram */
    printf("INIT_ACK_TGM Read Init Ack TGM.\n");
    break;
  case SS_TGM:			/* read state response tram */
    printf("SS_TGM Read Sick State TGM.\n");
    break;
  case BMACK_TGM:		/* change operating mode tram */
    printf("BMACK_TGM Change Operating Mode TGM. -- ");
    if (! (*tram->data))
      puts ("Ok.");
    else
      puts("Error.");
    break;
  case NACK_TGM:		/* Error tram */
    printf("NACK_TGM No Ack TGM.\n");
    break;
  case LMSKFG_TGM:		/* read config tram */
    printf("LMSKFG_TGM Read Config TGM.\n");
    break;
  case LMSKFGACK_TGM:		/* read config response tram */
    printf("LMSKFGACK_TGM Response To Config TGM. -- ");
    if (! (*tram->data)) 
      puts ("Error");
    else
      puts ("OK");
    break;
  case VARDEFACK_TGM:		/* change scan value tram */
    printf("VARDEFACK_TGM Response To Change Scanning Values TGM. -- ");
    if (! (*tram->data)) 
      puts ("Error");
    else
      puts ("OK");
    break;
  case ERR_TGM:
    printf("ERR_TGM Response To Error Get TGM. -- %d\n", (int) (*tram->data & 0x7f));
    break;
  case TYPE_TGM:
    printf("TYPE_TGM Response To Get Type TGM. -- ");
    tram->data[tram->len] = 0;
    puts (tram->data);
    break;
  default:
    printf("unimplemented command: 0x%2x\n", (int)tram->cmd);
  }
    /* --- */
    puts ("--------------------------------------------------------------------------------");
}

/*
 * Interpret trams
 */
static long
sick_lib_parse_tram (SICK_LIB_DATA_TRAM_STR *tram, int n, int *state)
{
  char *p = NULL, *start, *data_adrs, *head;
  unsigned short len, crc, c_crc;
  char cmd;
  long nl = n;

  /* set state */
  *state = OK;

  /* reposition start */
  head = tram->data;

  /* Find tram header = STX + STATION_ADR */
  start = sick_lib_find_tram_header (tram->data, n);
  nl -= (long)start - (long)head;

  /* check it */
  if (!start) {
    return 0;
  }

  /* skip STX + STATION_ADR -- position p on length field*/
  p = start + 2;

  /* Get data length */
  len = SICK_LIB_SHORT_FORMAT (*(unsigned short *)p);
    
  /* Check consistency of tram length */
  if (len > SICK_LIB_DATA_LENGTH_MAX || len < 1) {
    logMsg ("Sick Parser: Bad tram len\nRestart interpretation!\n");
    *state = ERROR;
    return (long) p - (long) head;
  }
    
  /* Check rest of buffer */
  if (len + 4 > nl - 2) {
    /* try a new one */
    char *tmp = sick_lib_find_tram_header_3 (start + 2, nl - 2);
    if (!tmp)
      return 0;

    logMsg ("Sick Parser: TGM was cut!\n");
    *state = ERROR;
    return (long) tmp - (long) head;
  }

  /* Get cmd */
  p += 2;
  cmd = *p;
  
  /* check command */
  if (sick_lib_check_command (cmd) != OK) {
    *state = ERROR;
    return (long) p - (long) head;
  }

  /* set command */
  tram->cmd = cmd;
  
  /* Keep data address */
  data_adrs = p + 1;
  p += len;

  /* Get sent check sum */
  crc = *(unsigned short *)p;
  
  /* recompute local check sum */
  c_crc = build_crc16 (start, (long)p - (long)start);
  
  /* compare computed and sent check sums */
  if (c_crc != crc) {
    logMsg ("Sick Parser: Received CRC = 0x%x\n", (int)crc);
    logMsg ("Sick Parser: Computed CRC = 0x%x\n", (int)c_crc);
    logMsg ("Sick Parser: Restart interpretation!\n");
    *state = ERROR;
    p += 2;
    return (long) p - (long) head;
  } 
      
  /* good tram */
  tram->len = (long)p - (long)data_adrs - 1;
  tram->status = *(p - 1);

  if (tram->len)
    memmove (tram->data, data_adrs, tram->len);

  /* Jump check sum */
  p += 2; 

  return (long) p - (long) head;
}

/* Masque des bits significatifs d'une mesure en fonction des flags demande's  
 * voir commentaire de l'enum SICK_LIB_MEASURE_FLAGS dans sick_lib.h */
static unsigned short measure_mask[] =
 { 0x1fff, 0x1fff, 0x1fff, 0x3fff, 0x3fff, 0x7fff, 0x7fff };

/*
 * Update measurements
 */
static void
sick_lib_update_measure (SICK_LIB_STR *sick, double *d, double *theta)
{
  unsigned short *p = (unsigned short *)sick->tram.data;
  double th, dth, scale;
  int i;
  unsigned short mask = measure_mask[sick->config.measure_flag];
  
  /* set theta min */
  th = (sick->rcv.range == SICK_LIB_SCAN_RANGE_180) ? -PI/2. : -DEG_TO_RAD(100)/2.;

  /* set dth */
  dth = ((sick->rcv.resolution == SICK_LIB_SCAN_RESOLUTION_100) ?
	 DEG_TO_RAD(1) : DEG_TO_RAD (.5));
  
  /* set scale */
  switch (sick->rcv.measure_unit) {
  case SICK_LIB_MEASURE_UNIT_CM:
    scale = .01;		/* in cm */
    break;
  case SICK_LIB_MEASURE_UNIT_10_CM:
    scale = .1;			/* in 10cm */
    break;
  default:
    scale = .001;		/* in mm */
  } /* switch */

  /* ranges */
  sick->measures.theta_min = th;
  sick->measures.dtheta = dth;

  /* get points number */
  sick->measures.n = (int) (SICK_LIB_SHORT_FORMAT (*p) & 0x3fff);
  p++;

  /* check */
  if (!d) 
    return;
  if (theta) {
    /* fill distances and angles */
    for (i = 0; i < sick->measures.n; i++, p++, th += dth) {
      d[i] = scale * (double) (SICK_LIB_SHORT_FORMAT(*p) & mask);
      theta[i] = th;
    } /* for */
  } else {
    /* fill distances */
    for (i = 0; i < sick->measures.n; i++, p++)
      d[i] = scale * (double) (SICK_LIB_SHORT_FORMAT(*p) & mask);
  }
}

/*
 * Find STX + STATION_ADR
 */
static char *
sick_lib_find_tram_header (char *buf, int size)
{
  char *p = buf;
  
  while (p < buf + size) {
    if (*p == SICK_LIB_STX && *(p + 1) == STATION_ADR) 
      return p;
    p++;
  } /* while */

  return NULL;
}


/*
 * Find ACK + STX + STATION_ADR
 */
static char *
sick_lib_find_tram_header_3 (char *buf, int size)
{
  char *p = buf;
  
  while (p < buf + size) {
    if (*p == SICK_LIB_ACK && *(p + 1) == SICK_LIB_STX && *(p + 2) == STATION_ADR) 
      return p;
    p++;
  } /* while */

  return NULL;
}



/*********************/
/*      TESTS        */
/*********************/

static POSTER_ID sick_lib_poster = NULL;
#define SICK_LIB_POSTER_NAME "sickPoster"
int sick_lib_read_task_stop = FALSE;
SICK_LIB_STR *sick_lib_id;

/*
 * Sick Read Task
 */
STATUS 
sick_lib_read_task (char *dev, int rate)
{
  int size;

  /* Init of sick */
  if (!(sick_lib_id = sick_lib_init (dev, rate))) {
    h2printErrno (0);
    return ERROR;
  }
  
  /* Create Poster */
  if (posterCreate (SICK_LIB_POSTER_NAME, SICK_LIB_DATA_LENGTH_MAX, 
		    &sick_lib_poster) != OK) {
    h2printErrno(errnoGet());
    return ERROR;
  }
    
  /* While sick alive */
  while (sick_lib_read_task_stop == FALSE) {
    /* check trams */
    size = sick_lib_read_entry_tram (sick_lib_id);

    if (size > 0) {
      printf ("sick read %d\n", size);
      /* Show tram */
      sick_lib_tram_show (&sick_lib_id->tram);
      if (sick_lib_is_measure_tram(sick_lib_id)) {
	  /* write poster */
	  sick_lib_write_poster(&sick_lib_id->tram);
      }
      if (sick_lib_update_sick_str (sick_lib_id, 0, 0) != OK)
	fprintf (stderr, "\t->Error in Sick Logic!\n");
      
    }
    /* wait */
    taskDelay (10);

  } /* while */

  printf ("Sick Read Task: Exit\n");

  posterDelete (sick_lib_poster);

  sick_lib_end (sick_lib_id);

  sick_lib_read_task_stop = 0;

  return OK;
}

/*
 * Test poster write
 */

static STATUS 
sick_lib_write_poster (SICK_LIB_DATA_TRAM_STR *tram)
{
  if (sick_lib_poster) {
    /* Write tram */
    if (posterWrite(sick_lib_poster, 0, tram->data, SICK_LIB_DATA_LENGTH_MAX) 
	!= SICK_LIB_DATA_LENGTH_MAX) {
	h2printErrno (errnoGet());
	taskSuspend(0);
    }
    return OK;
  }
  return ERROR;
}


/*
 * Nothing
 */

int
sick_lib_find_short (short data, int ad1, int length)
{
  long i;
  char h, l;
  
  h = (char) (data >> 8);
  l = (char) (0x00ff & data);

  if (length == 0) length = 0x400;
  
  for (i = ad1; (i - ad1) < length ; i++) {
    if (h == *(char *)i && l == *(char *)(i+1)) {
      printf("adresse 0x%lx\n", i);
      return i;
    }
  }
  return 0;
}

/*
 * Function used in simulation mode only (see sick_lib/simul/)
 */
STATUS sick_simul_read_env(char *pipo) {
  printf("sick : you are not in simulation mode !\n");
  return ERROR;
}
